a {
    color: #cccccc;
    font-weight: bold;
}


pre {
    margin:0px;
    opacity: 1;
}

.markdown-cell {
    margin-left: 40px;
    font-family:    Arial, Helvetica, sans-serif
}

.nei-output-div {
    overflow: hidden;
    width: 100%;
    font-family: Courier New;
    margin-top: 3px; 
    margin-left:10px; 
    /* background-color: rgba(256, 256, 256, 0.05);*/
}
